/*  nanowasp, an emulator for the microbee 128k
 *  Copyright (C) 2000,2001  David G. Churchill <froods@alphalink.com.au>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 *  Main nanowasp Module
 *
 *
 *  Provides init, main loop, and deinit.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <allegro.h>


#include "nanowasp.h"
#include "z80.h"
#include "vdu.h"
#include "crtc.h"
#include "memmap.h"
#include "keyb.h"
#include "fdc.h"
#include "mz80.h"
#include "gui.h"



static int init(void);
static int deinit(void);
static int reset(void);


FILE *debuglog;
int  debuglvl = 0;


int z80_cycles = 0;



static int init(void)
{
   if (debuglvl > 0)
   {
      debuglog = fopen("nwlog.txt", "w");
      if (debuglog == NULL)
      {
         fprintf(stderr, "unable to write nwlog.txt\n");
         return -1;
      }

      fprintf(debuglog, "debug level = %i\n", debuglvl);
   }


   allegro_init();   /* PORTME allegro */

   if (gui_init() != 0)
      return -1;

   textprintf(screen, font, 0, 200, 161,
              "+ z80 core\n");
   if (z80_init() != 0)
      return -1;

   textprintf(screen, font, 0, 210, 161,
              "+ memory mapper\n");
   if (memmap_init() != 0)
      return -1;

   textprintf(screen, font, 0, 220, 161,
              "+ crt controller\n");
   if (crtc_init() != 0)
      return -1;

   textprintf(screen, font, 0, 230, 161,
              "+ keyboard\n");
   if (keyb_init() != 0)
      return -1;

   textprintf(screen, font, 0, 240, 161,
              "+ vdu and graphics hardware\n");
   if (vdu_init() != 0)
      return -1;

   textprintf(screen, font, 0, 250, 161,
              "+ fdc\n");
   if (fdc_init() != 0)
      return -1;


   textprintf(screen, font, 0, 300, 161,
              ". . . press space to continue . . .");

   while (!key[KEY_SPACE]);

   return 0;
}



/* depends on short circuit evaluation */
static int deinit(void)
{
   if (debuglvl > 0)
      fclose(debuglog);


   return fdc_deinit() ||
          vdu_deinit() ||
          keyb_deinit() ||
          crtc_deinit() ||
          memmap_deinit() ||
          z80_deinit() ||
          gui_deinit();
}



static int reset(void)
{
   return z80_reset() ||
          memmap_reset() ||
          crtc_reset() ||
          keyb_reset() ||
          vdu_reset() ||
          fdc_reset();
}


int main(int argc, char *argv[])
{
   int opt;
   int done = 0;
   time_t starttime;
   int delay = 0, speed = 0;



   while ((opt = getopt(argc, argv, "d")) != -1)
   {
      switch(opt)
      {
         case 'd':  debuglvl++;  break;
         default:   exit(1);     break;
      }
   }


   if (init() != 0)
   {
      fprintf(stderr, "Fatal error during initialisation.  Exiting...\n");
      exit(1);
   }



   reset();

   /* main loop */
   starttime = time(NULL)-1;
   while (!done)
   {
      mz80exec(100000);
      redraw_screen();
      usleep(delay);


      z80_cycles += mz80GetElapsedTicks(TRUE);
      speed = z80_cycles / (time(NULL) - starttime);


      if (speed > SPEED)
         delay += SPEEDCORRECT;
      else
      {
         delay -= SPEEDCORRECT;
         if (delay < 0)
            delay = 0;
      }

      textprintf(screen, font, 10, 500, 80,
                 "speed: %i  delay: %i  cycles: %i  pc: 0x%04x       ",
                 speed, delay, z80_cycles, z80_getpc());

      if (key[KEY_F12])         /* PORTME allegro */
         done = 1;
      if (key[KEY_F11])
         reset();

      while (key[KEY_PAUSE]);
   }


   if (deinit() != 0)
   {
      fprintf(stderr, "Error while de-initialising.  Hrrmm... Oh well.\n");
      exit(1);
   }


   printf("last speed: %i kHz  last delay: %i usec\n",
          (int)(speed / 1000), delay);

   return 0;
}

